# -*- coding: utf8 -*-

#-----------------------------------------
# This is a part of Collège 3D project.
# Author:       Pascal Peter
# Copyright:    (C) 2003-2016 Pascal Peter
# Licence:      GNU General Public Licence version 3
# Website:      http://pascal.peter.free.fr/
# Email:        pascal.peter at free.fr
#-----------------------------------------

"""
DESCRIPTION :
    Execution d'un jeu.
"""

# **************************************************
#       IMPORTATION DES MODULES UTILES
# **************************************************
import os
import random
import time
import math
import mathutils

import bge

import utils, input_controller, game_controller



# **************************************************
#       CONTRÔLES DU JEU
# **************************************************
controller = None
scene = None
camera = None
player = None
gameName = None



# **************************************************
#       FONCTIONS UTILES
# **************************************************
def postDrawTexts():
    """
    affichage des textes.
    """
    if scene.objects['messagePlane'].visible:
        if bge.logic.globalDict['GAME']['state'] != 'game_loop':
            return
        for text in ('messageText', 'blue', 'red'):
            params = bge.logic.globalDict['TEXTS'][text]
            utils.print2D(params[0], params[4], params[2], params[3], params[5])
    else:
        utils.postDrawTexts('game_loop')

def playSound(soundName='clic'):
    if not(bge.logic.globalDict['MENU']['sounds']):
        return
    if soundName == 'clic':
        soundActuator = controller.actuators['Clic']
    elif soundName == 'over':
        soundActuator = controller.actuators['Over']
    elif soundName == 'jump':
        soundActuator = controller.actuators['Jump']
    else:
        soundActuator = None
    if soundActuator != None:
        controller.activate(soundActuator)

def playMusic(musicName='sonnerie'):
    if not(bge.logic.globalDict['MENU']['music']):
        return
    if musicName == 'sonnerie':
        musicActuator = controller.actuators['Sonnerie']
    else:
        musicActuator = None
    if musicActuator != None:
        controller.activate(musicActuator)

def keyboardManagement(inputState):
    """
    gestion du clavier
    """
    # touches spéciales :
    if inputState['QUIT']:
        if bge.logic.globalDict['GAMES'][gameName] != []:
            doQuitGame(nextScene='levels')
        elif gameName == 'LAN':
            doQuitGame(nextScene='lan_game')
        else:
            doQuitGame()
    elif inputState['MESSAGE']:
        game = bge.logic.globalDict['TEMP']['game']
        level = game.actualLevel
        mission = level.actualMission
        lang = bge.logic.globalDict['MENU']['lang']
        missionText = utils.tr('GAME_{0}'.format(gameName), mission.text, lang)
        utils.setText('missionText', text=missionText, time=5)
    elif inputState['CAMERA']:
        camera['viewCamera'] = not(camera['viewCamera'])
        mesh = scene.objects['mesh']
        if camera['viewCamera']:
            camera.localPosition = (0, 0, 2)
            mat_rot = mathutils.Matrix.Rotation(math.radians(90.0), 3, 'X')
            camera.localOrientation = mat_rot
            mesh.visible = False
        else:
            camera.localPosition = (0, -6, 3)
            mat_rot = mathutils.Matrix.Rotation(math.radians(75.0), 3, 'X')
            camera.localOrientation = mat_rot
            mesh.visible = True
    elif inputState['TEST']:
        doTest()
    elif inputState['ENTER']:
        # touche entrée pour fermer les messages :
        doMessage(True)

def characterManagement(inputState):
    """
    On teste les déplacements à chaque boucle.
    """
    if bge.logic.globalDict['GAME']['state'] != 'game_loop':
        return
    action = 'wait'
    isMoving = False
    fps = bge.logic.getLogicTicRate()
    if fps < 1:
        fps = 1

    # -------------------------------------------------------------
    # détection des déplacements demandés :
    # -------------------------------------------------------------
    UP = inputState['PLAYER_UP']
    DOWN = inputState['PLAYER_DOWN']
    JUMP = inputState['PLAYER_JUMP']
    turn = input_controller.calcTurn(inputState, player, fps)

    # -----------------------
    # calcul de l'avancée :
    # -----------------------
    maxAdvance = 10
    (advance, speedAdvance) = player['advance']
    if UP:
        action = 'walk'
        advance = 1
        if speedAdvance < 5:
            speedAdvance = 5
        else:
            speedAdvance += 2 / fps
            if speedAdvance > maxAdvance:
                speedAdvance = maxAdvance
    elif DOWN:
        action = 'walk'
        advance = - 1
        if speedAdvance < 5:
            speedAdvance = 5
        else:
            speedAdvance += 2 / fps
            if speedAdvance > maxAdvance:
                speedAdvance = maxAdvance
    elif speedAdvance > 0:
        action = 'walk'
        speedAdvance -= 30 / fps
        if speedAdvance < 0:
            speedAdvance = 0
            advance = 0
            action = 'wait'
    player['advance'] = (advance, speedAdvance)
    advance = advance * speedAdvance

    # ------------------
    # calcul du saut :
    # ------------------
    speedJump, maxJump = 1, 20
    (jumping, jump) = player['jump']
    if jumping:
        JUMP = False
    if JUMP:
        jumping = True
        action = 'jump'
        playSound('jump')
        jump = maxJump / fps
    if jumping:
        jump -= speedJump / fps
        if jump <= 0:
            jumping = False
            jump = 0
        else:
            action = 'jump'
    player['jump'] = (jumping, jump)

    # ---------------------------------------------------
    # on ne fait ce qui suit que s'il y a déplacement :
    # ---------------------------------------------------
    player['isMoving'] = False
    if (advance != 0) or (turn != 0) or (jump > 0):
        # calcul de up (pour monter les escaliers) :
        player['isMoving'] = True
        pente, up = 0, 0
        positionEx = player['positionEx']
        position = player.position
        deltaV = position[2] - positionEx[2]
        deltaH = math.hypot((positionEx[0] - position[0]), (positionEx[1] - position[1]))
        if deltaH > 0:
            pente = deltaV / deltaH
        if (pente > 0):
            if pente < 1:
                up = pente * advance
            elif pente < 5:
                up = advance
        player['positionEx'] = (player.position[0], player.position[1], player.position[2])
        # enfin on déplace le personnage :
        advance = advance * 0.01
        up = up * 0.01
        jump = jump * 2
        direction = player.orientation * mathutils.Vector((0, advance, up + jump))
        player['wrapper'].walkDirection = direction
        player.applyRotation([0, 0, turn], True)
        # correction en cas d'erreur 
        # (personnage en dehors des limites, sous le sol, etc) :
        worldLimites = ((-100, 100), (-100, 100))
        if player.position[0] < worldLimites[0][0]:
            player.position[0] = worldLimites[0][0] + 1
        elif player.position[0] > worldLimites[0][1]:
            player.position[0] = worldLimites[0][1] - 1
        if player.position[1] < worldLimites[1][0]:
            player.position[1] = worldLimites[1][0] + 1
        elif player.position[1] > worldLimites[1][1]:
            player.position[1] = worldLimites[1][1] - 1
        if player.position[2] < -5:
            player.position[2] = 20

    # -------------------------------------
    # la caméra tourne autour du player :
    # -------------------------------------
    if not(camera['viewCamera']):
        if (advance == 0) and (jump <= 0):
            speedTurn = 2
            turn = speedTurn / fps
            camera['angle'] += turn
            if camera['angle'] > 2 * math.pi:
                pivot_camera = scene.objects['pivot_camera']
                mat_rot = mathutils.Matrix.Rotation(0.0, 3, 'X')
                pivot_camera.localOrientation = mat_rot
                camera['angle'] = random.uniform(-25, -15)
            elif camera['angle'] > 0:
                pivot_camera = scene.objects['pivot_camera']
                pivot_camera.applyRotation([0, 0, turn], True)
        elif camera['angle'] > 0:
            speedTurn = 2
            turn = speedTurn / fps
            camera['angle'] += turn
            if camera['angle'] > 2 * math.pi:
                pivot_camera = scene.objects['pivot_camera']
                mat_rot = mathutils.Matrix.Rotation(0.0, 3, 'X')
                pivot_camera.localOrientation = mat_rot
                camera['angle'] = random.uniform(-25, -15)
            else:
                pivot_camera = scene.objects['pivot_camera']
                pivot_camera.applyRotation([0, 0, turn], True)
        else:
            camera['angle'] = -20

    # -------------------------------------
    # affichage éventuel du commentaire :
    # -------------------------------------
    position = player.position
    commentObject = player.rayCastTo((position[0], position[1], position[2] - 1), 2, 'comment')
    if commentObject != None:
        player['outside'] = False
        comment = commentObject['comment']
        comment = utils.tr('comments', comment, bge.logic.globalDict['MENU']['lang'])
        if comment != bge.logic.globalDict['TEMP']['lastComment']:
            bge.logic.globalDict['TEMP']['lastComment'] = comment
            utils.setText('textComment', text=comment, time=3)
    else:
        player['outside'] = True

    # -------------------------------------
    # mise à jour de l'action si besoin :
    # -------------------------------------
    if action != player['action']:
        actionData = utils.actions[action]
        player['action'] = action
        characterName = bge.logic.globalDict['MENU']['character']
        character = scene.objects[characterName]
        theAction = actionData[0]
        character.playAction(theAction, actionData[1], actionData[2], play_mode=actionData[3])

def scenarioManagement():
    if player['isMoving'] == False:
        return
    game = bge.logic.globalDict['TEMP']['game']
    level = game.actualLevel
    mission = level.actualMission
    if mission.cible == None:
        return

    try:
        cible = scene.objects[mission.cible[3]]
    except:
        utils.loadLib(mission.cible[2])
        cibleObjet = scene.objectsInactive[mission.cible[3]]
        cible = scene.addObject(cibleObjet, cibleObjet)
        cible.position = mission.cible[0]
    if player.getDistanceTo(cible) > mission.cible[1]:
        return

    try:
        player['hook'].endObject()
    except:
        pass
    try:
        player['head'].endObject()
    except:
        pass
    if mission.result != None:
        if not(mission.result in bge.logic.globalDict['GAME']['createdOjects']):
            utils.loadLib(mission.result[0])
            resultName = mission.result[1]
            resultObjet = scene.objectsInactive[resultName]
            result = scene.addObject(resultObjet, resultObjet)
            result.visible = True
            bge.logic.globalDict['GAME']['createdOjects'].append(mission.result)

    lang = bge.logic.globalDict['MENU']['lang']
    mission = level.chooseMission()
    if mission.noMission:
        # il n'y a plus de mission, donc le niveau est terminé.
        # on passe au niveau suivant :
        level = game.nextLevel()
        if level.noLevel:
            # il n'y a plus de niveau, donc le jeu est terminé :
            doQuitGame(nextScene=game.gameWon)
            return
        if level.name != '':
            # on retourne au menu des niveaux :
            doQuitGame(nextScene='levels')
            return
        if level.initialPosition != (-1, -1, -1):
            player.position = level.initialPosition
        mission = level.actualMission
        bge.logic.globalDict['GAME']['level'] = level.name
        if level.name != '':
            levelText = utils.tr('GAME_{0}'.format(gameName), level.name, lang)
            utils.setText('levelText', text=levelText, time=5)
        if level.weather != '':
            weather = utils.weatherDict[level.weather]
            doWeather(weather[0], weather[1])
    # si on arrive ici c'est qu'on est passé à la mission suivante
    if mission.time != None:
        player['deadline'] = mission.time[0]
    if mission.hook != None:
        try:
            hook = scene.objects['hook']
        except:
            hook = scene.objects['hook2']
        utils.loadLib(mission.hook[0])
        objet = scene.objectsInactive[mission.hook[1]]
        newObjet = scene.addObject(objet, objet)
        newObjet.position = hook.position
        newObjet.orientation = hook.orientation
        newObjet.setParent(hook)
        player['hook'] = newObjet
    if mission.head != None:
        try:
            head = scene.objects['head']
        except:
            head = scene.objects['head2']
        utils.loadLib(mission.head[0])
        objet = scene.objectsInactive[mission.head[1]]
        newObjet = scene.addObject(objet, objet)
        newObjet.position = head.position
        newObjet.orientation = head.orientation
        newObjet.setParent(head)
        player['head'] = newObjet
    if mission.cible != None:
        cible.position = mission.cible[0]
        cible.visible = False
        missionText = utils.tr('GAME_{0}'.format(gameName), mission.text, lang)
        utils.setText('missionText', text=missionText, time=5)

def doQuitGame(nextScene='menu'):
    if bge.logic.globalDict['GAME']['state'] == 'game_quit':
        return
    bge.logic.globalDict['GAME']['state'] = 'game_quit'
    # cas d'un jeu réseau :
    if gameName == 'LAN':
        bge.logic.globalDict['LAN']['sent_message'] = 'END'
        time.sleep(0.5)
    # on libère les fichiers chargés :
    #utils.freeAllLibs()
    utils.freeAllLibs(doAll=True)
    utils.clearGlobalDict()
    # on libère les textures temporaires :
    if gameName == 'Exhibition':
        for wall in utils.exhibitionWalls:
            utils.removeTexture(wall)
    # et on retourne au menu (ou à la scène demandée) :
    scene.replace(nextScene)

def doTest():
    player.position[2] = 0
    return
    if not('test' in player):
        player['test'] = False
    player['test'] = not(player['test'])
    if player['test']:
        print('TEST')
    else:
        print('FIN TEST')
    player.position[2] = 20
    doQuitGame(nextScene='expo')

def chooseWeather():
    """
    gestion des objets affichés en fonction de la saison
    """
    (objetName, total) = utils.chooseWeather()
    doWeather(objetName, total)

def doWeather(objetName='', total=0):
    """
    gestion des objets affichés en fonction de la saison
    """
    def doMist(value, start, end):
        if utils.BGE_VERSION > 273:
            scene.world.mistColor = [value, value, value]
            scene.world.mistStart = start
            scene.world.mistDistance = end
            scene.world.mistEnable = True
        else:
            bge.render.setMistColor([value, value, value])
            bge.render.setMistStart(start)
            bge.render.setMistEnd(end)
    # on supprime les précédents :
    for objet in player['season']['intemperies']:
        objet.endObject()
    player['season']['intemperies'] = []
    player['season']['intemperie'] = objetName
    if utils.BGE_VERSION > 273:
        scene.world.mistEnable = False
    else:
        bge.render.disableMist()
    # et on crée les intempéries :
    if objetName == '':
        return
    elif objetName == 'nuit':
        doMist(0, 3, total)
    elif objetName == 'brouillard':
        doMist(0.7, 3, total)
    else:
        utils.loadLib('objets/saisons')
        objet = scene.objectsInactive[objetName]
        for i in range(total):
            newObjet = scene.addObject(objet, objet)
            position = list(player.position)
            delta = (random.uniform(-5, 5), random.uniform(-5, 5), random.uniform(0, 10))
            position[0] = position[0] + delta[0]
            position[1] = position[1] + delta[1]
            position[2] = position[2] + delta[2]
            position = tuple(position)
            newObjet.position = position
            newObjet['delta'] = delta
            player['season']['intemperies'].append(newObjet)
        # on ajoute du brouillard :
        if objetName == 'pluie':
            value = 0.7
            doMist(0.7, 30, 40)
        elif objetName == 'flocon':
            doMist(0.7, 3, 20)

def doMessage(visible=None):
    """
    aaa
    """
    messagePlane = scene.objects['messagePlane']
    if visible == None:
        visible = messagePlane.visible
    if visible:
        messagePlane.visible = False
        utils.setText('messageText', text='')
        return
    messageText = utils.doMessage()
    if messageText != '':
        messagePlane.visible = True
        utils.setText('messageText', text=messageText, font=2, time=10)

def updateDetails():
    """
    met à jour les objets affichés (détails)
    """
    repere = scene.objects['repere']
    for detail in bge.logic.globalDict['TEMP']['details']:
        try:
            name = detail[0]
            repere.position = detail[1]
            rayon = detail[2]
            loaded = detail[3]
            visible = detail[4]
            objetName = name.split('/')[-1]
            distance = camera.getDistanceTo(repere)
            if distance < rayon:
                if not(loaded):
                    utils.loadLib(name)
                    detail[3] = True
                    detail[4] = True
                    objet = scene.objectsInactive[objetName]
                    newObjet = scene.addObject(objet, objet)
                    detail[5] = newObjet
                elif not(visible):
                    objet = scene.objectsInactive[objetName]
                    newObjet = scene.addObject(objet, objet)
                    detail[5] = newObjet
                    detail[4] = True
            elif visible:
                objet = detail[5]
                objet.endObject()
                detail[4] = False
        except:
            continue



# **************************************************
#       FONCTION PRINCIPALE
# **************************************************
def main():
    # ---------------------------------------------------------
    # RÉCUPÉRATION DE L'ÉTAT DU JEU
    # ---------------------------------------------------------
    # si on lance le fichier directement
    if not('GAME' in bge.logic.globalDict):
        bge.logic.globalDict['GAME'] = {}
    state = bge.logic.globalDict['GAME'].get('state', '')
    if not(state in ('game_first', 'game_loop', 'game_quit')):
        bge.logic.globalDict['GAME']['state'] = 'game_first'
    if bge.logic.globalDict['GAME']['state'] == 'game_quit':
        return

    # ---------------------------------------------------------
    # PREMIER LANCEMENT
    # ---------------------------------------------------------
    if bge.logic.globalDict['GAME']['state'] == 'game_first':
        # on masque la souris :
        bge.render.showMouse(False)
        # valeurs initiales (lues depuis le fichier de config si existe) :
        utils.loadSettings()
        global controller, scene, camera, player, gameName
        controller = bge.logic.getCurrentController()
        scene = bge.logic.getCurrentScene()
        camera = scene.active_camera
        player = scene.objects['player']
        player['wrapper'] = bge.constraints.getCharacter(player)
        gameName = bge.logic.globalDict['MENU']['gameName']

        # initialisation des timers :
        bge.logic.globalDict['TEMP']['actualTime'] = 0
        bge.logic.globalDict['TEMP']['actualPos'] = (0, 0)
        bge.logic.globalDict['TEMP']['timeSpecific'] = {4: False}

        # affichage des textes :
        scene.post_draw += [postDrawTexts]
        bge.logic.globalDict['TEXTS'] = {}
        utils.setText('levelText', color=utils.colors['orange'], width=5, position=(0.1, 0.2))
        utils.setText('missionText', color=utils.colors['blue'], width=4, position=(0.05, 0.3))
        utils.setText('textComment', color=utils.colors['violet'], width=3, position=(0.1, 0.9))
        utils.setText('textTimer', position=(0.02, 0.98))
        utils.setText('textInfos', position=(0.3, 0.98))
        utils.setText('messageText', width=2, position=(0.1, 0.7))
        utils.setText('blue', color=utils.colors['blue'], width=5, position=(0.05, 0.1))
        utils.setText('red', color=utils.colors['red'], width=5, position=(0.9, 0.1))
        bge.logic.globalDict['TEMP']['lastComment'] = ''

        # chargement des décors permanents :
        for objectName in utils.allwaysObjects['ALL_GAMES']:
            utils.loadLib(objectName)
        if gameName in utils.allwaysObjects:
            for objectName in utils.allwaysObjects[gameName]:
                utils.loadLib(objectName)
        # mise à jour des textures du jeu Exhibition :
        if gameName == 'Exhibition':
            utils.downloadFilesExhibition()
            for wall in utils.exhibitionWalls:
                wallObj = scene.objects[wall]
                utils.createTexture(wallObj, bge.logic.globalDict['downloadDir'], wall)

        # initialisation du jeu et du niveau :
        gameText = utils.tr('games', gameName, bge.logic.globalDict['MENU']['lang'])
        utils.setText('levelText', text=gameText, time=5)
        levelName = bge.logic.globalDict['GAME']['level']
        if not('game' in bge.logic.globalDict['TEMP']):
            game = game_controller.Game(gameName, levelName)
            bge.logic.globalDict['TEMP']['game'] = game
        game = bge.logic.globalDict['TEMP']['game']
        level = game.actualLevel
        mission = level.actualMission
        if mission.cible != None:
            utils.loadLib(mission.cible[2])
            cibleName = mission.cible[3]
            cibleObjet = scene.objectsInactive[cibleName]
            cible = scene.addObject(cibleObjet, cibleObjet)
            cible.position = mission.cible[0]
            if mission.time != None:
                cible.visible = (mission.time[0] <= mission.time[1])
                player['deadline'] = mission.time[0]

        # initialisation du personnage :
        if level.initialPosition != (-1, -1, -1):
            playerPosition = level.initialPosition
        elif game.initialPosition != None:
            playerPosition = game.initialPosition
        else:
            playerPosition = player.position
        player.position = playerPosition
        characterName = bge.logic.globalDict['MENU']['character']
        utils.loadChar(characterName)
        character = scene.objects[characterName]            
        character.position = (playerPosition[0], playerPosition[1], playerPosition[2] - 0.5)
        character.setParent(player)
        player['isMoving'] = False
        player['action'] = ''
        player['positionEx'] = (playerPosition[0], playerPosition[1], playerPosition[2])
        player['advance'] = (0, 0)
        player['jump'] = (False, 0)
        player['outside'] = True

        # le hook (objet porté par le personnage) :
        try:
            player['hook'].endObject()
        except:
            pass
        if mission.hook != None:
            try:
                hook = scene.objects['hook']
            except:
                hook = scene.objects['hook2']
            utils.loadLib(mission.hook[0])
            objet = scene.objectsInactive[mission.hook[1]]
            newObjet = scene.addObject(objet, objet)
            newObjet.position = hook.position
            newObjet.orientation = hook.orientation
            newObjet.setParent(hook)
            player['hook'] = newObjet
        # le chapeau du personnage :
        try:
            player['head'].endObject()
        except:
            pass
        if mission.head != None:
            try:
                head = scene.objects['head']
            except:
                head = scene.objects['head2']
            utils.loadLib(mission.head[0])
            objet = scene.objectsInactive[mission.head[1]]
            newObjet = scene.addObject(objet, objet)
            newObjet.position = head.position
            newObjet.orientation = head.orientation
            newObjet.setParent(head)
            player['head'] = newObjet
        # chargement des décors temporaires du jeu (objets créés) :
        for objectName in bge.logic.globalDict['GAME']['createdOjects']:
            utils.loadLib(objectName[0])
            objet = scene.objectsInactive[objectName[1]]
            newObjet = scene.addObject(objet, objet)

        # cas d'un jeu réseau :
        if gameName == 'LAN':
            import lan_client
            bge.logic.globalDict['LAN']['sent_message'] = ''
            bge.logic.globalDict['LAN']['received_message'] = ''
            host = bge.logic.globalDict['LAN']['host']
            user = bge.logic.globalDict['LAN']['user']
            password = bge.logic.globalDict['LAN']['password']
            ok = lan_client.main(host, utils.LAN_PORT)
            player['lan'] = [user, characterName, 'blue']
            player['lanThreads'] = {}
            player['lastMessages'] = ['', '', '', '']
            bge.logic.globalDict['LAN']['sent_message'] = 'CONNECTø{0}ø{1}'.format(user, characterName)
            bge.logic.globalDict['TEMP']['lanGame'] = ''

        # liste des détails du décor :
        bge.logic.globalDict['TEMP']['details'] = []
        for detail in utils.detailsObjects:
            bge.logic.globalDict['TEMP']['details'].append(
                [detail[0], detail[1], detail[2], False, False, None])
        updateDetails()
        # gestion du décor en fonction de la date :
        calendar = bge.logic.globalDict['SEASONS']['calendar']
        if calendar != None:
            utils.loadLib(calendar[0])
            calendarName = scene.objectsInactive[calendar[1]]
            calendarObjet = scene.addObject(calendarName, calendarName)
            calendarObjet.position = calendar[2]
            calendarObjet.visible = True
            if calendar[1] in utils.weatherDict:
                level.weather = calendar[1]
        # gestion de la saison :
        player['season'] = {'intemperie':'', 'intemperies':[], 'outside':True}
        if level.weather != '':
            weather = utils.weatherDict[level.weather]
            doWeather(weather[0], weather[1])
        elif not(gameName in ('LAN',)):
            chooseWeather()
        # musique :
        playMusic('sonnerie')
        # switch et rotation de la caméra :
        camera['viewCamera'] = False
        camera['angle'] = random.uniform(-5, -2)

        # initialisation terminée :
        player.position = playerPosition
        player['positionEx'] = (playerPosition[0], playerPosition[1], playerPosition[2])
        bge.logic.globalDict['TEMP']['initialTime'] = int(time.time())
        bge.logic.globalDict['GAME']['state'] = 'game_loop'

    # ---------------------------------------------------------
    # BOUCLE DES ÉVÉNEMENTS
    # ---------------------------------------------------------
    if bge.logic.globalDict['GAME']['state'] != 'game_loop':
        return
    inputState = input_controller.inputManagement('game')
    # touche PAUSE :
    if inputState['PAUSE']:
        bge.logic.globalDict['GAME']['pause'] = not(bge.logic.globalDict['GAME']['pause'])
        if bge.logic.globalDict['GAME']['pause']:
            message = utils.tr('other', 'GAME PAUSED', bge.logic.globalDict['MENU']['lang'])
        else:
            message = ''
        utils.setText('levelText', text=message, time=1)
    if bge.logic.globalDict['GAME']['pause']:
        return
    # pour faire un screencast :
    if inputState['SCREENSHOT']:
        utils.makeScreenshot()
    # calculs divers concernant les durées :
    actualTime = int(time.time() - bge.logic.globalDict['TEMP']['initialTime'])

    # gestion des touches spéciales :
    keyboardManagement(inputState)
    if bge.logic.globalDict['GAME']['state'] != 'game_loop':
        return

    # on teste les déplacements à chaque boucle :
    characterManagement(inputState)
    scenarioManagement()
    if bge.logic.globalDict['GAME']['state'] != 'game_loop':
        return

    # pour un jeu réseau, on envoie sa position et on gère les autres joueurs :
    if gameName == 'LAN':
        # message émis (sera envoyé après) :
        [user, characterName, color] = player['lan']
        sent_message = ''
        # message reçu :
        received_message = bge.logic.globalDict['LAN']['received_message']
        try:
            received_message = received_message.split('&')
            if len(received_message) > 1:
                # on vérifie la liste des joueurs :
                clientsText = received_message[1][8:]
                if player['lastMessages'][2] != clientsText:
                    player['lastMessages'][2] = clientsText
                    # on supprime les joueurs déconnectés
                    # et on met à jour les nombres de bleus et de rouges :
                    lanThreads, lanThreadsToDelete = [], []
                    blue, red = 0, 0
                    clients = clientsText.split('ø')
                    for client in clients:
                        clientDatas = client.split('#')
                        if len(clientDatas) > 2:
                            lanThreads.append(clientDatas[0])
                            colorOther = clientDatas[2]
                            if colorOther == 'blue':
                                blue += 1
                            elif colorOther == 'red':
                                red += 1
                    for threadName in player['lanThreads']:
                        if not(threadName in lanThreads):
                            lanThreadsToDelete.append(threadName)
                    for threadName in lanThreadsToDelete:
                        player['lanThreads'][threadName][2].endObject()
                        del player['lanThreads'][threadName]
                    utils.setText('blue', text='{0}'.format(blue))
                    utils.setText('red', text='{0}'.format(red))
                # on vérifie le dernier message du serveur :
                serverLast = received_message[2]
                if not(serverLast in ('', player['lastMessages'][3])):
                    player['lastMessages'][3] = serverLast
                    #print('serverLast :', serverLast)
                    received_message[0] = 'threadøSERVERø{0}'.format(serverLast)
            received_message = received_message[0]
            if not(received_message in ('', player['lastMessages'][1])):
                player['lastMessages'][1] = received_message
                received_message = received_message.split('ø')
                threadName = received_message[0]
                who = received_message[1]
                what = received_message[2]
                if who == 'SERVER':
                    # message du serveur
                    if what == 'WEATHER':
                        weather = utils.weatherDict[received_message[3]]
                        doWeather(weather[0], weather[1])
                    elif what == 'GAME':
                        print('GAME :', received_message[3:])
                        lanGame = received_message[3]
                        bge.logic.globalDict['TEMP']['lanGame'] = lanGame
                        lang = bge.logic.globalDict['MENU']['lang'] 
                        if lanGame == 'wolf':
                            wolf = received_message[4]
                            if wolf == '###':
                                wolf = user
                            if wolf == user:
                                message = utils.tr('GAME_LAN', 'You are the wolf!', lang)
                                player['lan'][2] = 'red'
                            else:
                                message = utils.tr('GAME_LAN', 'Wolf is:', lang)
                                message = '{0} {1}'.format(message, wolf)
                                player['lan'][2] = 'blue'
                            messagePlane = scene.objects['messagePlane']    
                            messagePlane.visible = True
                            utils.setText('messageText', text=message, font=2, time=10)
                        elif lanGame == 'zombies':
                            wolf = received_message[4]
                            if wolf == '###':
                                wolf = user
                            if wolf == user:
                                message = utils.tr('GAME_LAN', 'You are a zombie!', lang)
                                player['lan'][2] = 'red'
                            else:
                                message1 = utils.tr('GAME_LAN', 'has been zombified!', lang)
                                message2 = utils.tr(
                                    'GAME_LAN', 'The zombies are contagious.', lang)
                                message = '{0} {1}ø{2}'.format(wolf, message1, message2)
                            messagePlane = scene.objects['messagePlane']    
                            messagePlane.visible = True
                            utils.setText('messageText', text=message, font=2, time=10)
                    elif what == 'MESSAGE':
                        messageText = received_message[3]
                        messagePlane = scene.objects['messagePlane']    
                        messagePlane.visible = True
                        utils.setText('messageText', text=messageText, font=2, time=10)
                    elif what == 'LOOP':
                        # ICI : cas à vérifier
                        weather = utils.weatherDict[received_message[6][8:]]
                        doWeather(weather[0], weather[1])
                        lanGame = received_message[7][5:]
                        if bge.logic.globalDict['TEMP']['lanGame'] != lanGame:
                            bge.logic.globalDict['TEMP']['lanGame'] = lanGame
                            data = received_message[8][5:]
                            print('ICI :  GAME :', lanGame, 'DATA :', data)
                elif what == 'LOOP':
                    # message normal d'un autre joueur
                    characterOther = received_message[4]
                    colorOther = received_message[5]
                    colorValue = {'blue': [0, 0, 1, 1], 'red': [1, 0, 0, 1]}
                    m = received_message[6:]
                    position = (float(m[0]), float(m[1]), float(m[2]))
                    orientation = mathutils.Quaternion(
                        (float(m[3]), float(m[4]), float(m[5]), float(m[6]))).to_matrix()
                    if not(threadName in player['lanThreads']):
                        # nouveau joueur :
                        other = scene.objectsInactive['other']
                        newPlayer = scene.addObject(other, other)
                        newPlayer.position = position
                        newPlayer.localOrientation = orientation
                        player['lanThreads'][threadName] = [who, 
                            characterOther, newPlayer, actualTime, colorOther]
                        newPlayer.children['Cube_other'].visible = True
                        newPlayer.children['Cube_other'].color = colorValue[colorOther]
                        textPlayer = newPlayer.children['Cube_other'].children['Text']
                        textPlayer.text = who
                        textPlayer.resolution = 4.0
                    else:
                        # mise à jour du joueur :
                        player['lanThreads'][threadName][2].position = position
                        player['lanThreads'][threadName][2].localOrientation = orientation
                        player['lanThreads'][threadName][3] = actualTime
                        if colorOther != player['lanThreads'][threadName][4]:
                            player['lanThreads'][threadName][4] = colorOther
                            player['lanThreads'][threadName][2].children['Cube_other'].color = colorValue[colorOther]
                    doCollision = False
                    lanGame = bge.logic.globalDict['TEMP']['lanGame']
                    if color == 'red':
                        if player.getDistanceTo(player['lanThreads'][threadName][2]) < 1:
                            if lanGame in ('wolf', 'zombies'):
                                if not(scene.objects['messagePlane'].visible):
                                    if colorOther == 'blue':
                                        doCollision = True
                    if doCollision:
                        sent_message = 'COLLISIONø{0}ø{1}ø{2}ø{3}ø'.format(
                            user, characterName, color, who)
        except:
            #print('ERREUR received_message :', received_message)
            pass

        # on envoie enfin le message réseau :
        if sent_message == '':
            position = player.position
            orientation = player.orientation.to_quaternion()
            color = player['lan'][2]
            sent_message = 'LOOPø{0}ø{1}ø{2}ø{3}ø{4}ø{5}ø{6}ø{7}ø{8}ø{9}ø'.format(
                user, characterName, color,
                utils.doRound(position[0]), utils.doRound(position[1]), utils.doRound(position[2]),
                utils.doRound(orientation[0]), utils.doRound(orientation[1]), 
                utils.doRound(orientation[2]), utils.doRound(orientation[3]))
        if sent_message != player['lastMessages'][0]:
            bge.logic.globalDict['LAN']['sent_message'] = sent_message
            player['lastMessages'][0] = sent_message
        else:
            # pour envoyer un message en cas d'immobilité (voir plus loin) :
            sent_message = ''

    # la suite ne sera effectuée que toutes les secondes :
    if actualTime <= bge.logic.globalDict['TEMP']['actualTime']:
        return
    # mise à jour du actualTime :
    bge.logic.globalDict['TEMP']['actualTime'] = actualTime

    # mise à jour des textes :
    finishedTexts = utils.updateTexts()
    if 'messageText' in finishedTexts:
        scene.objects['messagePlane'].visible = False

    # infos affichées si demandé :
    if bge.logic.globalDict['MENU']['infos']:
        # calcul de la vitesse de déplacement :
        posEx = bge.logic.globalDict['TEMP']['actualPos']
        position = player.position
        bge.logic.globalDict['TEMP']['actualPos'] = (position[0], position[1])
        vitesse = round(math.hypot((posEx[0] - position[0]), (posEx[1] - position[1])), 1)
        # coordonnées :
        coords = '({0},{1},{2})'.format(
            round(position[0]), round(position[1]), utils.doRound(position[2] - 0.675, precision=1))
        # FPS :
        #fps = int(bge.logic.getLogicTicRate())
        fps = int(bge.logic.getAverageFrameRate())
        utils.setText('textInfos', text='{0}    {1} m/s    {2} fps'.format(coords, vitesse, fps),
            color=utils.colors['white'], width=3)

    # mise à jour des objets affichés et des intempéries (pluie, neige, etc)
    # (mais seulement après avoir affiché le message initial)
    outside = player['outside']
    if bge.logic.globalDict['TEMP']['timeSpecific'][4]:
        # on met à jour les objets affichés (détails) :
        updateDetails()
        # on met à jour les intempéries (pluie, neige, etc) :
        if player['season']['outside'] != outside:
            player['season']['outside'] = outside
            if player['season']['intemperie'] in ('pluie', 'flocon'):
                for objet in player['season']['intemperies']:
                    objet.visible = outside
        if outside:
            if player['season']['intemperie'] in ('pluie', 'flocon'):
                for objet in player['season']['intemperies']:
                    objet.position[0] = player.position[0] + objet['delta'][0]
                    objet.position[1] = player.position[1] + objet['delta'][1]
                    if objet.position[2] < player.position[2] - 2:
                        objet.position[2] = player.position[2] + objet['delta'][2]

    # gestion du scénario :
    game = bge.logic.globalDict['TEMP']['game']
    level = game.actualLevel
    mission = level.actualMission
    if (actualTime > 4) and not(bge.logic.globalDict['TEMP']['timeSpecific'][4]):
        bge.logic.globalDict['TEMP']['timeSpecific'][4] = True
        lang = bge.logic.globalDict['MENU']['lang']
        levelText = utils.tr('GAME_{0}'.format(gameName), level.name, lang)
        missionText = utils.tr('GAME_{0}'.format(gameName), mission.text, lang)
        utils.setText('levelText', text=levelText, time=5)
        utils.setText('missionText', text=missionText, time=5)
    if mission.time != None:
        player['deadline'] -= 1
        cible = scene.objects[mission.cible[3]]
        cible.visible = (player['deadline'] <= mission.time[1])
        utils.setText('textTimer', text='{0} s'.format(player['deadline']),
            color=utils.colors['white'], width=3)
        if player['deadline'] < 0:
            doQuitGame(nextScene='gameover')

    # la suite ne sera effectuée que toutes les 10 secondes :
    if (actualTime % 10) > 0:
        return

    # pour un jeu réseau, on envoie sa position régulièrement
    # (pour éviter qu'un joueur qui ne bouge pas ne soit pas détecté) :
    if gameName == 'LAN':
        if sent_message == '':
            [user, characterName, color] = player['lan']
            position = player.position
            orientation = player.orientation.to_quaternion()
            sent_message = 'LOOPø{0}ø{1}ø{2}ø{3}ø{4}ø{5}ø{6}ø{7}ø{8}ø{9}ø'.format(
                user, characterName, color,
                utils.doRound(position[0]), utils.doRound(position[1]), utils.doRound(position[2]),
                utils.doRound(orientation[0]), utils.doRound(orientation[1]), 
                utils.doRound(orientation[2]), utils.doRound(orientation[3]))
            bge.logic.globalDict['LAN']['sent_message'] = sent_message
            player['lastMessages'][0] = sent_message

    # on met à jour les intempéries (pluie, neige, etc) :
    if player['season']['intemperie'] in ('feuille', 'poisscaille'):
        for objet in player['season']['intemperies']:
            objet.position[0] = player.position[0] + random.uniform(-5, 5)
            objet.position[1] = player.position[1] + random.uniform(-5, 5)
            objet.position[2] = player.position[2] + random.uniform(10, 30)


