# -*- coding: utf8 -*-

#-----------------------------------------
# This is a part of Collège 3D project.
# Author:       Pascal Peter
# Copyright:    (C) 2003-2016 Pascal Peter
# Licence:      GNU General Public Licence version 3
# Website:      http://pascal.peter.free.fr/
# Email:        pascal.peter at free.fr
#-----------------------------------------

"""
DESCRIPTION :
    Gestion des jeux et niveaux.
    * un jeu est composé de niveaux
    * un niveau est composé de missions
"""

# **************************************************
#       IMPORTATION DES MODULES UTILES
# **************************************************
import random

import bge



# **************************************************
#       LES CLASSES UTILES (GAME, LEVEL, MISSION)
# **************************************************
class Game:
    """
    un jeu est composé de niveaux
        * name : le nom du jeu
        * levelName : le niveau à lancer
        * levels : la liste des niveaux
    """
    def __init__(self, name='', levelName=''):
        self.name = name
        if name in bge.logic.globalDict['GAMES']['INITIAL_POSITIONS']:
            self.initialPosition = random.choice(
                bge.logic.globalDict['GAMES']['INITIAL_POSITIONS'][name])
        else:
            self.initialPosition = None
        if name in bge.logic.globalDict['GAMES']['WON_SCENES']:
            self.gameWon = bge.logic.globalDict['GAMES']['WON_SCENES'][name]
        else:
            self.gameWon = 'gamewon'
        self.levels = []
        self.actualLevel = None
        self.actualLevelIndex = -1
        if name in bge.logic.globalDict['GAMES']:
            self.levels = bge.logic.globalDict['GAMES'][name]
        if levelName != '':
            levelIndex = self.levelNameToIndex(levelName)
            self.goLevel(levelIndex)
        else:
            self.nextLevel()

    def levelNameToIndex(self, levelName):
        levelIndex, i = 0, 0
        for level in self.levels:
            if level[0]['name'] == levelName:
                levelIndex = i
            i += 1
        return levelIndex

    def nextLevel(self):
        if self.actualLevelIndex > -1:
            # on marque le niveau actuel comme réussi :
            bge.logic.globalDict['GAMES'][self.name][self.actualLevelIndex][1] = True
        self.actualLevelIndex += 1
        if self.actualLevelIndex < len(self.levels):
            level = Level(self.levels[self.actualLevelIndex][0])
        else:
            level = Level()
        self.actualLevel = level
        return level

    def goLevel(self, levelIndex):
        self.actualLevelIndex = levelIndex
        if self.actualLevelIndex < len(self.levels):
            level = Level(self.levels[self.actualLevelIndex][0])
        else:
            level = Level()
        self.actualLevel = level
        return level

class Level:
    """
    un niveau est composé de missions
        * name : le nom du niveau ; si vide, c'est qu'il succède au précédent 
                et n'est pas un vrai niveau.
                Par exemple, dans le jeu "découvrir", chaque niveau est fait de 2 parties
                la première avec la cible visible, la seconde sans.
                Cela permet de sortir les missions au hasard tout en ne mélangeant pas
                les 2 parties du niveau.
        * initialPosition : une liste des positions initiales du joueur pour le niveau.
                S'il y en a plusieurs, la position est prise au hasard dans la liste.
        * missions : la liste des missions du niveau
        * random : indique si les missions sont effectuées au hasard ou dans l'ordre
        * weather : la météo du niveau
    """
    def __init__(self, params={}):
        self.params = params
        self.noLevel = (len(params) < 1)
        self.name = params.get('name', '')
        self.random = params.get('random', False)
        self.weather = params.get('weather', '')
        self.reInit()

    def reInit(self):
        positions = self.params.get('initialPositions', ((-1, -1, -1),))
        self.initialPosition = random.choice(positions)
        self.missions = self.params.get('missions', [])
        # test pour savoir s'il y a une ou plusieurs possibilités :
        if isinstance(self.missions, tuple):
            self.missions = random.choice(self.missions)
        self.actualMission = None
        self.chooseMission()

    def chooseMission(self):
        if len(self.missions) > 0 :
            if self.random :
                choice = random.choice(self.missions)
            else :
                choice = self.missions[0]
            self.actualMission = Mission(choice)
            self.missions.remove(choice)
        else:
            self.actualMission = Mission()
        return self.actualMission

class Mission:
    """
    une mission est un dictionnaire.
        * text : le texte affiché (ordre à executer par le joueur)
        * cible (si elle existe) : sa ou ses position(s), le rayon de détection, l'objet affiché
        * time (si le temps est limité) : sa valeur, on affiche la cible lorsqu'il reste
        * hook : l'objet porté à la main pendant la mission
        * head : l'objet porté sur la tête pendant la mission
        * result : l'objet créé à la fin
    """
    def __init__(self, params={}):
        self.noMission = (len(params) < 1)
        # le texte de la mission (ordre affiché) :
        self.text = params.get('text', '')
        # la cible :
        if 'cible' in params:
            cible = params['cible']
            # test pour savoir s'il y a une ou plusieurs positions :
            position = cible[0]
            if isinstance(cible[0][0], tuple):
                position = random.choice(cible[0])
            self.cible = (position, cible[1], cible[2], cible[3])
        else:
            self.cible = None
        # le reste :
        self.time = params.get('time', None)
        self.hook = params.get('hook', None)
        self.head = params.get('head', None)
        if self.head != None:
            if isinstance(self.head[0], tuple):
                self.head = random.choice(self.head)
        self.result = params.get('result', None)



# **************************************************
#       DES FONCTIONS UTILES
# **************************************************
def reloadMissions(gameName, levelName):
    # on recrée la liste des missions :
    import copy
    import scenarios
    index = -1
    levelIndex = -1
    nextIndex = -1
    for [level, ok] in bge.logic.globalDict['GAMES'][gameName]:
        index += 1
        if levelIndex < 0:
            if level['name'] == levelName:
                bge.logic.globalDict['GAMES'][gameName][index][0] = copy.deepcopy(
                    scenarios.levels[gameName][index])
                bge.logic.globalDict['GAMES'][gameName][index][1] = False
                levelIndex = index
        elif nextIndex < 0:
            if level['name'] == '':
                bge.logic.globalDict['GAMES'][gameName][index][0] = copy.deepcopy(
                    scenarios.levels[gameName][index])
                bge.logic.globalDict['GAMES'][gameName][index][1] = False
                levelIndex = index
            else:
                nextIndex = index


